import networkx as nx
from scipy.stats import hypergeom
from statsmodels.stats.multitest import fdrcorrection

import goenrich.export

def analyze(H, query, **kwargs):
    """ run enrichment analysis for query

    >>> O = goenrich.obo.graph('...')
    >>> background = goenrich.read.goa('...')
    >>> H = goenrich.enrich.set_background(O, background, ...)
    >>> df = goenrich.enrich.analyze(H, query, ...)

    :param H: Ontology graph after backgroud was set
    :param query: array like of ids
    :returns: pandas.DataFrame with results
    """
    options = {
            'node_filter' : lambda node : 'p' in node,
            'show' : 'top20',
            'inplace' : True
    }
    options.update(kwargs)
    if not options['inplace']:
        raise ValueError('goenrich.enrich.analyze needs inplace=False')
    pvalues = calculate_pvalues(H, query, **options)
    multiple_testing_correction(H, pvalues, **options)
    df = goenrich.export.to_frame(H, **options)
    if 'gvfile' in options:
        show = options['show']
        if show.startswith('top'):
            top = int(show.replace('top', ''))
            sig = df.sort('q').head(top).index
        else:
            raise NotImplementedError(show)
        goenrich.export.to_graphviz(H, sig, **options)
    return df
    
def set_background(O, df, entry_id, category_id, background_name='', inplace=False):
    """ Creates a copy of the ontology tree and
    propagates the background set through it

    >>> O = goenrich.obo.graph('...')
    >>> background = goenrich.read.goa('...')
    >>> H = goenrich.enrich.set_background(O, background, ...)
    
    :param O: ontology graph: generated by geonrich.obo.graph
    :param df: background data set: look in goenrich.read for parsers
    :param entry_id: protein or gene identifier column
    :param category_id: GO term column
    :param background_name: specify background annotation name
    :param inplace: force in-place annotation
    :returns H: the annotated ontology graph if inplace was not set
    """
    H = O if inplace else O.copy()
    M = len(df[entry_id].unique()) # total number of objects
    if inplace: # clean background attribute for changing backgrounds
        for n in H:
            node = H.node[n]
            node['background'] = set([])

    grouped = df.groupby(category_id)[entry_id]
    for term,entries in grouped:
        namespace = H.node[term]['namespace']
        root = H.graph['roots'][namespace]
        for path in nx.simple_paths.all_simple_paths(H, term, root):
            for n in path:
                node = H.node[n]
                node.setdefault('background', set()).update(entries)
                if 'M' not in node:
                    node['M'] = M
    H.graph['background'] = background_name
    if not inplace:
        return H

def calculate_pvalues(H, query, min_hit_size=2, min_category_size=3,
        max_category_size=500, max_category_depth=5,
        inplace=False, **kwargs):
    """ calculate pvalues for all categories in the graph
    
    :param H: ontology graph after background was set
    :param query: array_like of identifiers
    :param min_hit_size: minimum intersection size of query and category 
    :param min_category_size: categories smaller than this number are ignored
    :param max_category_size: categories larger than this number are ignored
    :param inplace: perform operations inplace or return copy
    :returns: dictionary of term : pvalue
    """
    G = H if inplace else H.copy()
    query_set = set(query)
    pvalues = {}
    N = len(query_set)
    for i in G:
        node = G.node[i]
        if inplace:
            # reset all query related attributes
            for attr in ['query', 'n', 'N', 'hits', 'x', 'p', 'q', 'significant']:
                if attr in node:
                    del node[attr]
        background = node.get('background', set([]))
        n = len(background)
        node['n'] = n
        hits = query_set.intersection(background)
        x = len(hits)
        depth = node.get('depth', -1) # depth might not be set due to malformed ontology
        if ((depth > max_category_depth)
            or (n < min_category_size)
            or (n > max_category_size)
            or (x < min_hit_size)):
            continue
        else:
            node['query'] = query_set
            node['N'] = N
            node['hits'] = hits
            node['x'] = x
            M, n = node['M'], node['n']
            p = hypergeom.sf(x, M, n, N)

            node['p'] = p
            pvalues[i] = p
    if inplace:
        return pvalues
    else:
        return G, pvalues


def multiple_testing_correction(G, pvalues, alpha=0.05,
        method='benjamini-hochberg', inplace=False, **kwargs):
    """ correct pvalues for multiple testing and add corrected `q` value
    
    :param G: ontology graph for which pvalues were calculated
    :param alpha: significance level default : 0.05
    :param method: multiple testing correction method [bonferroni|benjamini-hochberg]
    """
    H = G if inplace else G.copy()
    H.graph.update({ 'multiple-testing-correction': method, 'alpha' : alpha })
    if method == 'bonferroni':
        n = len(pvalues.values())
        for term,p in pvalues.items():
            node = H.node[term]
            q = p * n
            node['q'] = q
            node['significant'] = q < 0.05
    elif method == 'benjamini-hochberg':
        terms, ps = zip(*pvalues.items())
        rejs, qs = fdrcorrection(ps, alpha)
        for term, q, rej in zip(terms, qs, rejs):
            node = H.node[term]
            node['q'] = q
            node['significant'] = rej
    else:
        raise ValueError(method)
    if not inplace:
        return H



"""
tools for working with goslims
"""
import re
import networkx as nx

def add(O, goslim, nodes):
    """ add goslim to specified nodes in the ontology """
    for n in nodes:
        O.node[n].setdefault('subset', set()).add(goslim)

def subset(O, goslim):
    """ get the ontology-subset corresponding to the specified goslim

    :param goslim: specify a goslim which the ontolgy is filtered for
    """
    if goslim is not None and goslim not in O.graph['goslims']:
        raise ValueError('goslim {} is not part of the ontology definition'.format(goslim))
    goslim_nodes = [(n,data) for n,data in O.nodes(data=True)
            if goslim in data.get('subset', set())]
    G = O.subgraph(n for n,_ in goslim_nodes)
    for term, node in goslim_nodes:
        root = G.graph['roots'][node['namespace']]
        nodes = [n for path in nx.all_simple_paths(O, term, root)
                for n in path if goslim in O.node[n].get('subset', set())]
        G.add_edges_from(e for e in zip(nodes, nodes[1:]))
    return G

def read(filename):
    """ read .obo file assuming all 'id: GO:XXXXXXX' belong to the
    goslim

    :returns: list of go terms found in the file """
    findid = re.compile('^id: (GO:\d+)')
    with open(filename) as f:
        for line in f:
            m = findid.match(line)
            if m is not None:
                yield m.groups()[0]



import networkx as nx
import pandas as pd

import goenrich

def to_frame(G, node_filter = lambda node: True, **kwargs):
    """ export enrichment graph to pd.DataFrame

    :param node_filter: decision function based with access to the node data dictionary
    """
    return (pd.DataFrame([
        {k: v for d in [{'term':term}, node] for k,v in d.items()}
        for term, node in G.nodes(data=True) if node_filter(node)])
        .set_index('term'))

def to_graphviz(G, sig, gvfile, goslim=None, graph_label=None, **kwargs):
    """ export graph of signifcant findings to dot file.
    A png can be generated from the commandline using graphviz

    dot -Tpng filpath.dot > filepath.png

    :param sig: array_like of node lables to include in result graph
    :param gvfile: file or filepath
    :param goslim: contract graph and skip categories which are not part
        of the specified goslim
    :param graph_label: give custom graph label otherwise
        additional information will be printed.
        For empty label pass graph_label=''.
    """
    nodes = set([])
    for n in sig:
        namespace = G.node[n]['namespace']
        root = G.graph['roots'][namespace]
        for path in nx.simple_paths.all_simple_paths(G, n, root):
            if goslim is not None:
                start = n
                end = None
                for i, m in path[1:]:
                    if goslim in G.node[m]['subset']: # TODO
                        continue
            nodes.update(path)
    R = G.subgraph(nodes).reverse()
    for n in R:
        node = R.node[n]
        attr = {}
        attr['shape'] = 'record'
        if 'q' in node:
            attr['color'] = 'red' if node['significant'] else 'black'
            attr['label'] = """{name}
            {x} / {n} genes
            q = {q:.5f}""".format(name=node['name'], q=node['q'], x=node['x'], n=node['n'])
        else:
            attr['color'] = 'black'
            attr['label'] = """{name}""".format(name=node['name'])
        R.node[n] = attr

    A = nx.to_agraph(R)
    if graph_label is not None:
        A.graph_attr['label'] = graph_label
    else:
        A.graph_attr['label'] = 'multiple testing correction: {}\nat alpha={}'.format(
                    G.graph['multiple-testing-correction'], G.graph['alpha'])
    A.graph_attr['labelloc'] = 't'
    
    if hasattr(gvfile, 'write'):
        A.write(gvfile)
    else:
        with open(gvfile, 'w') as f:
            A.write(f)




import re
import os
import pickle
import networkx as nx

def add(G, id, name, namespace, is_a=None, subset=None, **kwargs):
    """ add term to ontology graph """
    if next(iter(kwargs.get('def', set([''])))).startswith('"OBSOLETE'):
        return 'obsolete'
    else:
        # assumed unique
        id = next(iter(id))
        name = next(iter(name)) 
        namespace = next(iter(namespace)) 
        if is_a is not None:
            for parent in is_a:
                G.add_edge(id, parent)
        attr_dict = {}
        if subset is not None:
            attr_dict['subset'] = subset
        G.add_node(id, name=name, namespace=namespace, attr_dict=attr_dict)
        if name == namespace:
            G.graph['roots'][name] = id
        return 'added'

def term(f, G):
    """ parse term starting after [Term] tag, until key: value pair is missed """
    attr = {}
    while True:
        line = f.readline()
        if line == '\n': # end of term found
            return add(G, **attr)
        else:
            # remove comments and split to `key: value`
            key,value = line.split('!')[0].strip().split(': ', 1)
            attr.setdefault(key, set()).add(value)

def graph(filename, verbose=False, ontology_cache=None):
    """ Generate ontology graph
    
    :param filename: .obo file name
    :param verbose: print parsing info
    :param ontology_cache: pickle graph to file for faster loading"""
    if ontology_cache is not None:
        if os.path.isfile(ontology_cache):
            with open(ontology_cache, 'rb') as f:
                return pickle.load(f)
    G = nx.DiGraph(roots={})
    i = 0
    obsolete = 0
    goslim_definition = re.compile('subsetdef: (goslim_\w+) "(.*)"')
    with open(filename) as f:
        while True:
            line = f.readline()
            match = re.match(goslim_definition, line)
            if match is not None:
                G.graph.setdefault('goslims', {}).update(dict([match.groups()]))
            # TODO regex
            if line == '[Term]\n': # start term
                i += 1
                if term(f, G) == 'obsolete':
                    obsolete += 1
            elif line == '': # entire file parsed
                break
    if verbose:
        print(i, 'terms parsed,', obsolete, 'obsolete ignored')

    Grev = G.reverse()
    for root in G.graph['roots'].values():
        G.node[root]['depth'] = 0
        for n, depth in nx.single_source_shortest_path_length(Grev, root).items():
            G.node[n]['depth'] = depth

    if ontology_cache is not None:
        with open(ontology_cache, 'wb') as f:
            pickle.dump(G, f)
    return G



"""
README.md example
"""

import goenrich

# build the ontology
O = goenrich.obo.graph('db/go-basic.obo')

# use all entrez geneid associations form gene2go as background
# use goenrich.read.goa('db/gene_association.goa_ref_human.gz') for uniprot
background = goenrich.read.gene2go('db/gene2go.gz')
G = goenrich.enrich.set_background(O, background, 'GeneID', 'GO_ID')

# extract some list of entries as example query
query = set(background['GeneID'].unique()[:20])

# for additional export to graphviz just specify the gvfile argument
# the show argument keeps the graph reasonably small
result = goenrich.enrich.analyze(G, query, gvfile='example.dot', show='top20')

# generate html
result[['name', 'x', 'p', 'q', 'namespace']].head().to_csv('example.csv')

# call to graphviz
import subprocess
subprocess.call(['dot', '-Tpng', 'example.dot', '-o', 'example.png'])


# goslim example
# dowload goslim from ftp://ftp.ebi.ac.uk/pub/databases/GO/goa/goslim/goslim_goa.obo
goa_slim = goenrich.goslim.read('db/goslim_goa.obo')
goenrich.goslim.add(G, 'goslim_goa', (n for n in goa_slim if n in G))
S = goenrich.goslim.subset(G, 'goslim_goa')
result_slim = goenrich.enrich.analyze(S, query, gvfile='example_slim.dot', show='top20')

# generate html
result_slim[['name', 'x', 'p', 'q', 'namespace']].head().to_csv('example_slim.csv')

# call to graphviz
subprocess.call(['dot', '-Tpng', 'example_slim.dot', '-o', 'example_slim.png'])
